

#include <ctype.h>
#include <iostream>
#include <sstream>
#include <string>
#include "LexicalAnalyzer.H"


using namespace v1processor;
using namespace std;


ostream &v1processor::operator << (ostream &os, const LexicalAnalyzerException &e) {
	os << "Lexical exception: " << e.message << " at line " << e.line << ", column " << e.column;
	return os;
}


const string Token::RESERVED_WORDS[Token::NUM_RESERVED_WORDS] = {
	"loadi", "load", "op", "assign", "add", "sub", "and", "or", "xor", "not", "inc", "dec", "slr", "sar", "sll", "addz", "addn", "store", "push", "pop", "j", "jz", "jn", "ra", "rb", "rc", "sp", "pc"
};


Token *Token::newIdInstance(string image, int line, int column) {
	Token *ret = NULL;
	for (int i = 0; i < NUM_RESERVED_WORDS; i++) {
		if (image == RESERVED_WORDS[i]) {
			ret = new Token(i, image, line, column);
			break;
		}
	}
	if (ret == NULL)
		ret = new Token(Token::TYPE_ID, image, line, column);
	return ret;
}


Token *Token::newNumberInstance(string image, int line, int column) {
	istringstream iss(image);
	int v;
	iss >> v;
	return new Token(v, line, column);
}


ostream &v1processor::operator << (ostream &os, const Token &v) {
	os << "type=";
	if (v.type == Token::TYPE_ID)
		os << "ID (" << v.image << ")";
	else if (v.type == Token::TYPE_COMMA)
		os << "COMMA (,)";
	else if (v.type == Token::TYPE_COLON)
		os << "COLON (:)";
	else if (v.type == Token::TYPE_NUMBER)
		os << "NUMBER (" << v.number << ")";
	else if (v.type == Token::TYPE_EOF)
		os << "EOF";
	else
		os << "RESERVED WORD #" << v.type << " (" << v.image << ")";
	return os;
}


LexicalAnalyzer::LexicalAnalyzer(istream &is) {
	this->is = &is;
	this->status = STATUS_IDLE;
	this->line = 1;
	this->column = 1;
	this->oss.str("");
}


Token *LexicalAnalyzer::getNextToken() {
	//cout << "getNextToken: " << endl;
	char c;
	while (this->tokenQueue.empty() && (this->status != STATUS_STOP)) {
		if (this->is->get(c)) {
			//cout << "c = '" << c << "', status = " << this->status << endl;
			if (c == '\n') {
				this->line++;
				this->column = 1;
			}
			if (this->status == STATUS_IDLE) {
				if (c == ',')
					this->tokenQueue.push(new Token(Token::TYPE_COMMA, ",", this->line, this->column));
				else if (c == ':')
					this->tokenQueue.push(new Token(Token::TYPE_COLON, ":", this->line, this->column));
				else if (c == ';')
					this->status = STATUS_LINE_COMMENT;
				else if (c == '/')
					this->status = STATUS_PRE_MULTILINE_COMMENT;
				else if (isalpha(c) || (c == '_')) {
					this->status = STATUS_ID;
					this->oss.str("");
					this->oss << c;
				}
				else if (isdigit(c) || (c == '-')) {
					this->status = STATUS_NUMBER;
					this->oss.str("");
					this->oss << c;
				}
			}
			else if (this->status == STATUS_LINE_COMMENT) {
				if ((c == '\n') || (c == '\r'))
					this->status = STATUS_IDLE;
			}
			else if (this->status == STATUS_PRE_MULTILINE_COMMENT) {
				if (c == '*')
					this->status = STATUS_MULTILINE_COMMENT;
				else if (c == '/')
					this->status = STATUS_LINE_COMMENT;
				else
					throw LexicalAnalyzerException("invalid comment", this->line, this->column);
			}
			else if (this->status == STATUS_MULTILINE_COMMENT) {
				if (c == '*')
					this->status = STATUS_POST_MULTILINE_COMMENT;
			}
			else if (this->status == STATUS_POST_MULTILINE_COMMENT) {
				if (c == '/')
					this->status = STATUS_IDLE;
				else
					this->status = STATUS_MULTILINE_COMMENT;
			}
			else if (this->status == STATUS_ID) {
				if (isalnum(c) || (c == '_'))
					this->oss << c;
				else {
					this->tokenQueue.push(Token::newIdInstance(this->oss.str(), this->line, this->column));
					if (c == ',') {
						this->tokenQueue.push(new Token(Token::TYPE_COMMA, ",", this->line, this->column));
						this->status = STATUS_IDLE;
					}
					else if (c == ':') {
						this->tokenQueue.push(new Token(Token::TYPE_COLON, ":", this->line, this->column));
						this->status = STATUS_IDLE;
					}
					else if (c == ';')
						this->status = STATUS_LINE_COMMENT;
					else if (c == '/')
						this->status = STATUS_PRE_MULTILINE_COMMENT;
					else if (isspace(c))
						this->status = STATUS_IDLE;
					else
						throw LexicalAnalyzerException("expected letter, number, comma, colon or comment start sequence", this->line, this->column);
				}
			}
			else if (this->status == STATUS_NUMBER) {
				if (isdigit(c))
					this->oss << c;
				else {
					this->tokenQueue.push(Token::newNumberInstance(this->oss.str(), this->line, this->column));
					if (c == ',') {
						this->tokenQueue.push(new Token(Token::TYPE_COMMA, ",", this->line, this->column));
						this->status = STATUS_IDLE;
					}
					else if (c == ':') {
						this->tokenQueue.push(new Token(Token::TYPE_COLON, ":", this->line, this->column));
						this->status = STATUS_IDLE;
					}
					else if (c == ';')
						this->status = STATUS_LINE_COMMENT;
					else if (c == '/')
						this->status = STATUS_PRE_MULTILINE_COMMENT;
					else if (isspace(c))
						this->status = STATUS_IDLE;
					else
						throw LexicalAnalyzerException("expected number, comma, colon or comment start sequence", this->line, this->column);
				}
			}
			this->column++;
		}
		else {
			this->status = STATUS_STOP;
			this->tokenQueue.push(new Token(Token::TYPE_EOF, "", this->line, this->column));
		}
	}
	return this->tokenQueue.front();
}


void LexicalAnalyzer::consumeNextToken() {
	if (!this->tokenQueue.empty())
		this->tokenQueue.pop();
}

